import random
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
from .chunk_power_spectrum import power_spectrum
from .red_noise import create_red_noise, theoretical_red_noise_power_spectrum, calc_a

class RedNoiseAnalysis:
    """
    Performing red noise analysis on a given signal.
    
    Parameters
    ----------
    signal : array_like (time,)
        Input signal (1D array).
    
    lag : int, optional
        Lag at which to calculate the regression coefficient. Default is 1.
        
    chunk_size : int, optional
        Chunk size. Default is 1825, 5 years for daily data.
        
    red_noise_simulate_length : int, optional
        Length of the simulated red noise signal. Default is 365000, 1000 years for daily data.
    
    Public Attributes
    -----------------
    signal : array_like (time,)
        Input signal (1D array).
    
    sp : ndarray (n_chunks, size)
        Power spectrum of the input signal.
    
    freq : ndarray (size,)
        Frequencies corresponding to the power spectrum of the input signal.
    
    freq_theo : ndarray (size,)
        Frequencies corresponding to the theoretical red noise power spectrum.
    
    sp_theo : ndarray (size,)
        Theoretical red noise power spectrum calculated according to the input signal.
        
    sp_red : ndarray (n_red_noise_ens, n_chunks, size)
        Power spectrum of the simulated red noise.
    
    freq_red : ndarray (size,)
        Frequencies corresponding to the power spectrum of the simulated red noise.
    """
    def __init__(self, signal, lag = 1, chunk_size = 1825, n_red_noise_ens = 1000, red_noise_simulate_length = 365000):
        self.signal           = signal
        self._lag             = lag
        self._chunk_size      = chunk_size
        self._n_red_ens       = n_red_noise_ens
        self._simulate_length = red_noise_simulate_length
        
        self.sp   = None
        self.freq = None
        
        self._a         = None
        self.freq_theo  = None
        self.sp_theo    = None
        self._red_noise = None
        self.sp_red     = None
        self.freq_red   = None
    
    def _create_red_noise_sample(self):
        """
        Bootstrapping chunks of red noise samples from the simulated red noise.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        red : ndarray (chunk_size,)
            Chunks of red noise samples.
        """
        pos  = random.choice(np.arange(self._simulate_length - len(self.signal)))
        red  = self._red_noise[pos:pos+len(self.signal)]
        red /= red.std()
        return red
        
    def get_power_spectrum(self):
        """
        Calculating power spectrum of the input signal and corresponding red noise.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        # Signal power spectrum
        print("Calculating power spectrum of the input signal...")
        self.freq, self.sp = power_spectrum(self.signal, self._chunk_size)
        
        # Theoretical red noise power spectrum
        print("Calculating theoretical red noise power spectrum...")
        self._a                      = calc_a(self.signal, self._lag)
        self.freq_theo, self.sp_theo = theoretical_red_noise_power_spectrum(self.signal, self._lag, self._chunk_size)
        
        # Simulated red noise power spectrum
        print("Calculating simulated red noise power spectrum...")
        self._red_noise = create_red_noise(self._a, self._simulate_length)
        self.sp_red     = np.zeros((self._n_red_ens,) + self.sp.shape)
        for _ in range(self._n_red_ens):
            red = self._create_red_noise_sample() # Bootstrapping
            self.freq_red, self.sp_red[_] = power_spectrum(red, self._chunk_size)
    
    def plot_power_spectrum(self, title = "Power Spectrum of Signal and Corresponding Red Noise", fig_name = "red_noise_analysis.png"):
        """
        Plotting power spectrum of the input signal and corresponding red noise.
        
        Parameters
        ----------
        fig_name : str, optional
            Name of the output figure file. Default is "red_noise_analysis.png".
        
        Returns
        -------
        None
        
        Note that get_power_spectrum() should be called before this method.
        The plot will be saved as a PNG file with the given name.
        This method is only for fast visualization. If you have specific plotting needs, calling this method is not suggested.
        """
        print("Plotting power spectrum...")
        print("The plot will be saved as a PNG file with the given name.")
        print("This method is only for fast visualization. If you have specific plotting needs, calling this method is not suggested.")
        sp_red_ens_mean  = self.sp_red.mean(axis = 1).mean(axis = 0)
        sp_red_ens_std   = self.sp_red.mean(axis = 1).std(axis = 0)
        sp_conf_ens_high = stats.norm.interval(0.95, sp_red_ens_mean, sp_red_ens_std)[1]
        sp_conf_ens_low  = stats.norm.interval(0.95, sp_red_ens_mean, sp_red_ens_std)[0]
        
        plt.figure(figsize=(10, 5))
        plt.plot(self.freq, self.sp.mean(axis = 0) / np.sum(self.sp.mean(axis = 0)), color = "k", linewidth = 1, label = "Signal")
        plt.plot(self.freq_red, sp_red_ens_mean / np.sum(sp_red_ens_mean), color = "dimgrey", linestyle = "--", label = "Simulated Red Noise")
        plt.fill_between(self.freq_red, sp_conf_ens_low / np.sum(sp_red_ens_mean), sp_conf_ens_high / np.sum(sp_red_ens_mean), color = "grey", alpha = 0.5)
        plt.plot(self.freq_theo, self.sp_theo / np.sum(self.sp_theo), color = "k", linestyle = ":", label = "Theoretical Red Noise")
        plt.legend(fontsize = 14)
        plt.xscale("log")
        plt.xlabel("frequency [timestep$^{-1}$]", fontsize = 14)
        plt.xticks(fontsize = 14)
        plt.ylabel("Explained Variance", fontsize = 14)
        plt.yticks(fontsize = 14)
        plt.title(title, fontsize = 16)
        plt.savefig(fig_name, dpi = 300)
        plt.show()


import random
import numpy as np
from numpy.fft import fft, fftfreq
from scipy.linalg import logm, expm

# lead-lag regression coefficient
def calc_a(signal, lag = 1):
    """
    Calculating the lead-lag regression coefficient at specified lag.
    
    Parameters
    ----------
    signal : array_like (time,)
        Input signal (1D array).
    
    lag : int, optional
        Lag at which to calculate the regression coefficient. Default is 1.
        Note that the lag can be greater than 1.
        
    Returns
    -------
    a : ndarray (1,)
        Lead-lag regression coefficient at the specified lag.
        If the given lag is greater than 1, the lag-th root of the regression coefficient matrix is returned.
    """
    x0 = signal[:-lag]
    x1 = signal[lag:]
    x0 = np.reshape(x0, (1, len(x0)))
    x1 = np.reshape(x1, (1, len(x1)))
    c0 = np.matmul(x0, x0.T)
    ct = np.matmul(x1, x0.T)
    a  = np.matmul(ct, np.linalg.inv(c0))
    a  = np.squeeze(expm(logm(a) / lag)) # take the lag-th root to get the regression coefficient at lag 1
    return a

# red noise model
def _time_integral(x: float, a: float, e: float):
    """
    Time integration of red noise model.
    
    Parameters
    ----------
    x : float
        Current value of the signal.
    
    a : float
        Lead-lag regression coefficient.
    
    e : float
        Noise term.
        
    Returns
    -------
    x_new : float
        New value of the signal after time integration.
    """
    return np.dot(a, x) + e

def _calc_noise(a, size = 1000):
    """
    Calculating noise term for red noise model.
    
    Parameters
    ----------
    a : float
        Lead-lag regression coefficient.
    
    size : int, optional
        Size of the white noise to be generated. Default is 1000.
        
    Returns
    -------
    noise : float
        Noise term for the red noise model.
    """
    sigma       = np.sqrt(1 - a**2)
    white_noise = np.random.normal(0, sigma, size = size)
    noise       = np.random.choice(white_noise)
    return noise

# Creating red noise, both theoretical and simulated
def theoretical_red_noise_power_spectrum(signal, lag = 1, chunk_size = 1825):
    """
    Creating theoretical red noise power spectrum according to given signal.
    
    Parameters
    ----------
    signal : array_like
        Input signal (1D array).
        
    lag : int, optional
        Lag at which to calculate the regression coefficient. Default is 1.
    
    chunk_size : int, optional
        Chunk size. Default is 1825, 5 years for daily data.
    
    Returns
    -------
    freq : ndarray (size,)
        Frequencies corresponding to the power spectrum.
    
    sp : ndarray (size,)
        Theoretical red noise power spectrum of the input signal.
    """
    a    = calc_a(signal, lag)
    size = np.floor(chunk_size / 2).astype(int)
    freq = fftfreq(chunk_size)
    sp   = (1 - a**2) / (1 + a**2 - 2 * a * np.cos(2 * np.pi * freq))
    return freq[:size], sp.real[:size]

def create_red_noise(a, simulate_length = 365000):
    """
    Creating simulated red noise according to given lead-lag regression coefficient.
    
    Parameters
    ----------
    a : float
        Lead-lag regression coefficient.
        
    simulate_length : int, optional
        Length of the simulated red noise signal. Default is 365000, 1000 years for daily data.
        
    Returns
    -------
    red : ndarray (simulate_length,)
        Simulated red noise signal.
    """
    red    = np.zeros(simulate_length)
    red[0] = np.random.normal(0, 1)
    for i in range(simulate_length - 1):
        noise    = _calc_noise(a)
        red[i+1] = _time_integral(red[i], a, noise)
    return red


import numpy as np
from numpy.fft import fft, fftfreq

def _separate_chunks(signal, chunk_size = 1825):
    """
    Separating signal into chunks.
    
    Parameters
    ----------
    signal : array_like (time,)
        Input signal (1D array).
        
    chunk_size : int, optional
        Chunk size. Default is 1825, 5 years for daily data.
        
    Returns
    -------
    chunks : ndarray (n_chunks, chunk_size)
        List containing separated chunks of the input signal.
    """
    if len(signal) < chunk_size:
        raise ValueError("Signal length must be at least as large as chunk_size.")
    
    chunk_length = chunk_size
    num_chunks   = len(signal) // chunk_length
    chunks = []
    for i in range(num_chunks):
        start = i * chunk_length
        end   = start + chunk_length
        chunks.append(signal[start:end])
    return np.array(chunks)

def _calc_power_spectrum(signal):
    """
    Calculating power spectrum
    
    Parameters
    ----------
    signal : array_like (time,)
        Input signal (1D array).
        
    Returns
    -------
    freq : ndarray (size,)
        Frequencies corresponding to the power spectrum.
    
    sp : ndarray (size,)
        Power spectrum of the input signal.
    """
    size = np.floor(len(signal) / 2).astype(int)
    ck   = fft(signal)
    freq = fftfreq(len(signal))
    sp   = 2 * ck * ck.conj() / len(signal)**2
    return freq[:size], sp.real[:size]

def power_spectrum(signal, chunk_size = 1825):
    """
    Calculating averaged power spectrum over chunks.
    
    Parameters
    ----------
    signal : array_like (time,)
        Input signal (1D array).
        
    chunk_size : int, optional
        Chunk size. Default is 1825, 5 years for daily data.
        
    Returns
    -------
    freq : ndarray (size,)
        Frequencies corresponding to the power spectrum.
    
    all_sp : ndarray (n_chunks, size)
        All power spectrum for each chunk of the input signal.
    """
    chunks = _separate_chunks(signal, chunk_size)
    all_sp = []
    
    for chunk in chunks:
        freq, sp = _calc_power_spectrum(chunk)
        all_sp.append(sp)
    all_sp = np.array(all_sp)
    
    return freq, all_sp


import pytest
import numpy as np
from numpy.fft import fft, fftfreq
from scipy.linalg import logm, expm
from pyRedNoise.red_noise import calc_a, create_red_noise
from .fixture import rmm_data

def test_calc_a():
    
    # testing data
    a = np.array([0.95 ** i for i in range(100)])
    
    # lag 1
    a1 = calc_a(a, lag = 1)
    assert a1 == pytest.approx(0.95, rel = 1e-2)
    
    # lag 5
    a5 = calc_a(a, lag = 5)
    assert a5 == pytest.approx(0.95, rel = 1e-2)

def test_red_noise(rmm_data):
    rmm1   = rmm_data[0]
    a_rmm1 = calc_a(rmm1, lag = 1)
    red    = create_red_noise(a_rmm1)
    a_red  = calc_a(red, lag = 1)
    assert a_red == pytest.approx(a_rmm1, rel = 1e-2)


import pytest
import numpy as np
from numpy.fft import fft, fftfreq
from pyRedNoise.chunk_power_spectrum import _separate_chunks, _calc_power_spectrum, power_spectrum

def test_separate_chunks():
    
    # test if ValueError is raised for too short signal
    short_signal = np.random.rand(1000)
    with pytest.raises(ValueError):
        _separate_chunks(short_signal, chunk_size=1825)
    
    # test normal case
    signal = np.random.rand(5000)
    chunks = _separate_chunks(signal, chunk_size=1825)
    assert chunks.shape == (2, 1825)
    
def test_calc_power_spectrum():
    
    # testing data
    x = np.linspace(0, 2 * np.pi, 5000)
    y = 2 * np.sin(x) + 0.2 * np.random.rand(5000)
    
    # test the output size
    freq, sp = _calc_power_spectrum(y)
    assert len(freq) == 2500
    assert len(sp)   == 2500
    
    # test the power spectrum values
    assert np.amax(sp) == pytest.approx(2**2/2, rel = 1e-2)
    
def test_power_spectrum():
    # testing data
    x = np.linspace(0, 2 * np.pi, 5000)
    y = 2 * np.sin(x) + 0.2 * np.random.rand(5000)
    
    freq, sp = power_spectrum(y, chunk_size = 1000)
    
    # test the output size
    assert freq.shape == (500,)
    assert sp.shape   == (5, 500,)


import pytest
import numpy as np
from pyRedNoise.RedNoiseAnalysis import RedNoiseAnalysis
from .fixture import rmm_data

def test_RedNoiseAnalysis():
    # testing data
    x = np.linspace(0, 2 * np.pi, 10000)
    y = 2 * np.sin(x) + 0.2 * np.random.rand(10000)
    
    rna = RedNoiseAnalysis(y, lag = 1, chunk_size = 2500, n_red_noise_ens = 100, red_noise_simulate_length = 365000)
    rna.get_power_spectrum()
    
    # test if attributes are correctly assigned
    assert rna.signal.shape    == (10000,)
    assert rna.sp.shape        == (4, 1250)
    assert rna.freq.shape      == (1250,)
    assert rna.freq_theo.shape == (1250,)
    assert rna.sp_theo.shape   == (1250,)
    assert rna.sp_red.shape    == (100, 4, 1250)
    assert rna.freq_red.shape  == (1250,)

def test_plotting(rmm_data):
    rmm1 = rmm_data[0]
    rna  = RedNoiseAnalysis(rmm1, lag = 1, chunk_size = 1825, n_red_noise_ens = 100, red_noise_simulate_length = 365000)
    rna.get_power_spectrum()
    # test plotting functions
    try:
        rna.plot_power_spectrum()
    except Exception as e:
        pytest.fail(f"Plotting methods raised an exception: {e}")